import unittest

import pandas as pd

from testy_quick.handlers import BaseComparer


class DataframeIndexedComparer(BaseComparer):
    def __init__(self,index_cols, **kwargs):
        self.index_cols = index_cols

    def assert_same(self, expected, actual, test_case: unittest.TestCase):
        expected_df = expected.set_index(self.index_cols).sort_index()
        actual_df = actual.set_index(self.index_cols).sort_index()
        pd.testing.assert_frame_equal(expected_df, actual_df, check_dtype=False)


import os

import pandas as pd

from . import settings
from .handlers import BaseReader, BaseAnswer
from .structures import TestyRunner


def read_input(complete_file_path, var_name, reader_key):
    """

    :param complete_file_path:
    :param var_name:
    :param reader_key:
    :return:
    """
    reader = BaseReader.get_handler(reader_key)
    ans = reader.read(complete_file_path, var_name)
    return ans


def run_fct(f, case):
    args, kwargs = _extract_inputs(case)
    ans = f(*args, **kwargs)
    return ans

def _reverse_join(path):
    head,tail=os.path.split(path)
    if head:
        return _reverse_join(head)+[tail]
    return [tail]
def _shorted_path(path):
    l=_reverse_join(path)
    i=1
    while i<len(l):
        if i==0:
            i+=1
            continue
        if l[i]=="..":
            del (l[i - 1])
            del (l[i - 1])
            i-=1
        else:
            i+=1
    return os.path.join(*l)



def _extract_inputs(case):
    folder = os.path.join(settings.BASE_DATA_FOLDER, case)
    reader_key = settings.INPUT_META_READER[0]
    input_meta_file = os.path.join(folder, settings.INPUT_META_FILE_NAME)
    input_meta: pd.DataFrame = read_input(input_meta_file, settings.INPUT_META_READER[1], reader_key)
    assert not input_meta[settings.INPUT_META_VAR_COL].duplicated().any()
    args = list()
    kwargs = dict()
    for _, r in input_meta.iterrows():
        var_name = r[settings.INPUT_META_VAR_COL]
        complete_file_path = os.path.join(folder, r[settings.INPUT_META_FILE_COL])
        complete_file_path=_shorted_path(complete_file_path)
        input_datum = read_input(
            complete_file_path=complete_file_path,
            var_name=var_name,
            reader_key=r[settings.INPUT_META_READER_COL]
        )
        if r[settings.INPUT_META_NAMED_COL]:
            kwargs[var_name] = input_datum
        else:
            args.append(input_datum)

    return args, kwargs


def _extract_outputs(case):
    folder = os.path.join(settings.BASE_DATA_FOLDER, case)
    reader_key = settings.OUTPUT_META_READER[0]
    output_meta_file = os.path.join(folder, settings.OUTPUT_META_FILE_NAME)
    input_meta: pd.DataFrame = read_input(output_meta_file, settings.OUTPUT_META_READER[1], reader_key)
    ans = list()
    for _, r in input_meta.iterrows():
        complete_file_path = os.path.join(folder, r[settings.OUTPUT_META_FILE_COL])
        complete_file_path = _shorted_path(complete_file_path)
        var_name = r[settings.OUTPUT_META_VAR_COL]
        handler = BaseAnswer.get_handler(r[settings.OUTPUT_META_READER_COL])
        output_datum = handler.read(complete_file_path, var_name)
        ans.append((var_name, output_datum, handler))
    return ans


def create_test(fct_to_test, case, case_name) -> TestyRunner:
    args, kwargs = _extract_inputs(case)
    run_fct_c = lambda: fct_to_test(*args, **kwargs)
    outputs = _extract_outputs(case)
    ans = TestyRunner(run_function=run_fct_c, check_functions=outputs, case_name=case_name)
    return ans



from typing import Dict, List

import pandas as pd


class _Registered:
    __registered: Dict[str, "_Registered"] = dict()

    def __new__(cls, *args, **kwargs):
        if "name" not in kwargs:
            raise TypeError("Constructor of a registered object must have a keyword 'name' argument as registry key. "
                            f"Solution: Modify the initializer function of {cls} and/or the call.")
        name = kwargs["name"]
        if not isinstance(name, str) or not name:
            raise TypeError("Name must be a non-empty string.")
        if name in cls.__registered:
            raise KeyError(f"A reader with the name {name} already exists. "
                           f"Use a different name form {cls.available_handlers()}. "
                           f"FYI: {cls.handlers_df()}")
        x = super().__new__(cls)
        cls.__registered[name] = x
        return x

    @classmethod
    def get_handler(cls, name: str) -> "_Registered":
        if name not in cls.__registered:
            raise KeyError(
                f"No handler named {name}. Available handlers: {cls.available_handlers()}. Call {cls}.handlers_df() for full info.")
        ans = cls.__registered[name]
        if not isinstance(ans, cls):
            raise TypeError(f"Handler {name} is of type {ans.__class__}, not {cls}")
        return ans

    @classmethod
    def available_handlers(cls) -> List[str]:
        return list(cls.__registered.keys())

    @classmethod
    def handlers_df(cls):
        ans = pd.DataFrame(columns=["name", "class", "object", "description"])
        for n, o in cls.__registered.items():
            d = {"name": n, "class": str(o.__class__), "object": str(o), "description": o.__doc__}
            ans = ans.append(d, ignore_index=True)
        return ans



from .base_reader import BaseReader

import pandas as pd


class CsvReader(BaseReader):
    def read(self, complete_file_name, var_name):
        df = pd.read_csv(complete_file_name)
        return df



import os
import shutil
#remove dist
dirpath = os.path.join('dist')
if os.path.exists(dirpath) and os.path.isdir(dirpath):
    shutil.rmtree(dirpath)

#upgrade_version
def f1(data):
    i=0
    for l in data:
        if l.startswith("__version__"):
            vf=l.split("=")[1].strip().strip('"').split(".")
            v=int(vf[2])+1
            vf[2]=str(v)
            vf=".".join(vf)
            return i,vf
        i+=1
    raise Exception()
filename=os.path.join("testy_quick","__init__.py")
with open(filename,"r") as f:
    data=f.readlines()
i,v=f1(data)
data[i]=f'__version__ = "{v}"\n'
with open(filename,"w") as f:
    f.writelines(data)

#build
from setuptools import setup
setup(script_args=["sdist","bdist_wheel"])

#deploy
from twine.cli import dispatch
dispatch(["upload","dist/*","-u",os.getenv('pip_id'),"-p",os.getenv('pip_password')])




import inspect
import unittest
from functools import partial
from typing import Iterable, List, Dict

from testy_quick.functions import create_test
from testy_quick.structures import TestyRunner, TestyCase


class SimpleTester():
    def __init__(self, test_cases: Iterable[TestyCase]):
        self.test_cases = test_cases

    def create_tests(self) -> Dict[str, TestyRunner]:
        ans = dict()
        for test_case in self.test_cases:
            for case_folder in test_case.cases:
                case_name = test_case.name + "_" + case_folder
                t = create_test(test_case.function, case_folder, case_name)
                ans[case_name]=t
        return ans


def _setup(self: unittest.TestCase, run_fct, outputs: List[str]):
    answer = run_fct()
    answer_dict = dict()
    if len(outputs) > 1:
        self.assertEqual(len(outputs), len(answer))
        for name, ans in zip(outputs, answer):
            answer_dict[name] = ans
    elif len(outputs) == 1:
        answer_dict[outputs[0]] = answer
    else:
        self.assertIsNone(answer)
    self.__class__.answer = answer_dict


def _compare(self, comparer, value, name):
    ans = comparer.assert_same(value, self.answer[name], self)
    return ans


def create_tests_here(tests: Iterable[TestyRunner], module=None):
    if module is None:
        frm = inspect.stack()[1]
        module = inspect.getmodule(frm[0])
    module = module.__dict__
    for runner in tests:
        _add_test(module, runner)


def _add_test(module, runner):
    if runner.case_name in module:
        raise KeyError(f"{runner.case_name} already in module.")
    test_class_dict = dict()
    test_class_dict["setup_fct"] = lambda self: _setup(self, runner.function_to_run,
                                                       [t[0] for t in runner.check_functions])
    test_class_dict["test_0_run_function"] = lambda self: self.setup_fct()
    for name, value, comparer in runner.check_functions:
        f1 = lambda self, c=comparer, n=name, v=value: c.assert_same(v, self.answer[n], self)
        f = partial(_compare, comparer=comparer, value=value, name=name)
        # test_class_dict["test_"+name] = f #ToDo: figure out why this does not work
        test_class_dict["test_" + name] = f1
    cls = type(runner.case_name, (unittest.TestCase,), test_class_dict)
    module[runner.case_name] = cls



import pandas as pd

from testy_quick.handlers import BaseReader


class CsvArgsReader(BaseReader):
    def __init__(self, name, **read_kwargs):
        self.read_kwargs = read_kwargs

    def read(self, complete_file_name, var_name):
        df = pd.read_csv(complete_file_name, **self.read_kwargs)
        return df


from . import BaseAnswer, CsvArgsReader, DataframeIndexedComparer


class IndexedDataframeAnswer(BaseAnswer, CsvArgsReader, DataframeIndexedComparer):
    def __init__(self, index_cols, name, **read_kwargs):
        CsvArgsReader.__init__(self, name=name, **read_kwargs)
        DataframeIndexedComparer.__init__(self, index_cols, name=name)



from abc import ABC, abstractmethod

from testy_quick.handlers._registered import _Registered


class BaseReader(_Registered, ABC):
    """
    metaclass to control readers.
    """

    @abstractmethod
    def read(self, complete_file_name, var_name):
        print("called BaseReader.read")



import unittest
import pandas as pd

from . import BaseComparer


class DataframeDefaultComparer(BaseComparer):
    def assert_same(self, expected, actual, test_case: unittest.TestCase):
        pd.testing.assert_frame_equal(expected,actual,check_dtype=False)



from typing import Callable, Iterable, List, Tuple, Any

from testy_quick.handlers.base_comparer import BaseComparer


class TestyCase:
    def __init__(self,function:Callable,cases:Iterable[str],short_name:str=None):
        self.function=function
        self.cases=cases
        if short_name:
            self.name=short_name
        else:
            self.name=function.__str__()

class TestyRunner:
    def __init__(self,run_function,check_functions:List[Tuple[str,Any,BaseComparer]],case_name):
        self.function_to_run= run_function
        self.check_functions:List[Tuple[str,Any,BaseComparer]]=check_functions
        self.case_name=case_name







import unittest
from abc import ABC, abstractmethod

from testy_quick.handlers._registered import _Registered


class BaseComparer(_Registered, ABC):
    @abstractmethod
    def assert_same(self,expected,actual,test_case:unittest.TestCase):
        pass



import os

BASE_DATA_FOLDER = os.path.join("testy", "data")

INPUT_META_FILE_NAME = "input_meta.csv"
INPUT_META_READER = ("csv", "")
INPUT_META_VAR_COL = "var_name"
INPUT_META_FILE_COL = "file_name"
INPUT_META_READER_COL = "reader_name"
INPUT_META_NAMED_COL = "is_named"

OUTPUT_META_FILE_NAME = "output_meta.csv"
OUTPUT_META_READER = ("csv", "")
OUTPUT_META_VAR_COL = "var_name"
OUTPUT_META_FILE_COL = "file_name"
OUTPUT_META_READER_COL = "handler_name"



import unittest

from . import BaseComparer


class DefaultComparer(BaseComparer):
    def assert_same(self, expected, actual, test_case: unittest.TestCase):
        test_case.assertEqual(expected, actual)



from .base_reader import BaseReader
from .base_comparer import BaseComparer
from .csv_reader import CsvReader
from .base_answer import BaseAnswer
from .dataframe_default_comparer import DataframeDefaultComparer
from .csv_default_answer import CsvDefaultAnswer
from .default_comparer import DefaultComparer
from .json_reader import JsonReader
from .default_json_answer import DefaultJsonAnswer
from .csv_args_reader import CsvArgsReader
from .dataframe_indexed_comparer import DataframeIndexedComparer
from .indexed_dataframe_answer import IndexedDataframeAnswer
from .json_file_reader import JsonFileReader

default_csv_answer = CsvDefaultAnswer(name="csv")
default_csv_reader = default_csv_answer
default_json_answer = DefaultJsonAnswer(name="json")
default_json_reader = default_json_answer
default_comparer = default_json_answer
default_json_file_reader = JsonFileReader(name="json_file")



import json

from .base_reader import BaseReader


class JsonReader(BaseReader):
    def read(self, complete_file_name, var_name):
        d = json.load(open(complete_file_name, "r"))
        return d[var_name]
